﻿local L = AceLibrary("AceLocale-2.2"):new("sRaidFrames")
local BS = LibStub:GetLibrary("LibBabble-Spell-3.0"):GetLookupTable()
local Aura = AceLibrary("SpecialEvents-Aura-2.0")
local roster = AceLibrary("Roster-2.1") 
local HealComm = LibStub:GetLibrary("LibHealComm-3.0", true)

local media = AceLibrary("SharedMedia-1.0")
local mType = media.MediaType and media.MediaType.STATUSBAR or "statusbar"

media:Register(mType, "Otravi", "Interface\\AddOns\\sRaidFrames\\textures\\otravi")
media:Register(mType, "Smooth", "Interface\\AddOns\\sRaidFrames\\textures\\smooth")
media:Register(mType, "Striped", "Interface\\AddOns\\sRaidFrames\\textures\\striped")
media:Register(mType, "BantoBar", "Interface\\AddOns\\sRaidFrames\\textures\\bantobar")

sRaidFrames = AceLibrary("AceAddon-2.0"):new(
	"AceDB-2.0",
	"AceEvent-2.0",
	"AceConsole-2.0",
	"FuBarPlugin-2.0"
)

local sRaidFrames = sRaidFrames

sRaidFrames.hasIcon = "Interface\\Icons\\INV_Helmet_06"
sRaidFrames.defaultMinimapPosition = 180
sRaidFrames.cannotDetachTooltip = true
sRaidFrames.hasNoColor = true
sRaidFrames.clickableTooltip = true
sRaidFrames.hideWithoutStandby = true
sRaidFrames.independentProfile = true

local function NormalizeSpellName(name)
	if ( GetLocale() == "koKR" ) then	
		return name:gsub(" ", "")
	else
		return name:gsub("[^%a]", "")
	end
end

function sRaidFrames:OnInitialize()

	self:RegisterDB("sRaidFramesDB")

	self:RegisterDefaults("profile", {
		Locked				= false,
		HealthFormat	= 'percent',
		Invert 				= false,
		Scale					= 1,
		Border				= true,
		Texture				= "Otravi",
		BuffType			= "debuffs",
		ShowOnlyDispellable	= 1,
		BackgroundColor		= {r = 0.32, g = 0.50, b = 0.70, a = 0.90},
		BorderColor		= {r = 1, g = 1, b = 1, a = 1},
		HealthTextColor	= {r = 1, g = 1, b = 1, a = 1},
		Growth				= "down",
		Spacing				= 0,
		ShowGroupTitles		= true,
		UnitTooltipMethod		= "notincombat",
		BuffTooltipMethod = "always",
		DebuffTooltipMethod = "always",
		BuffFilter			= {},
		DebuffFilter		= {},
		DebuffWhitelist = {},
		PowerFilter			= {[0] = true,[1] = true,[2] = true,[3] = true},
		RangeCheck 			= false,
		RangeLimit			= 40,
		RangeFrequency	= 2,
		RangeAlpha 			= 0.5,
		ReadyCheck			= true,
		AggroCheck			= false,
		HighlightTarget	= false,
		HighlightHeals	= true,
		HighlightHealsSelf = true,
		HighlightDebuffs = "onlyself",
		DebuffColors		= { ["Curse"] = { ["r"] = 1, ["g"] = 0, ["b"] = 0.75, ["a"] = 0.5},
												["Magic"]    = { ["r"] = 1, ["g"] = 0, ["b"] = 0, ["a"] = 0.5},
												["Disease"]  = { ["r"] = 1, ["g"] = 1, ["b"] = 0, ["a"] = 0.5},
												["Poison"]   = { ["r"] = 0, ["g"] = 0.5, ["b"] = 0, ["a"] = 0.5} },
		Layout					= "CTRA_WithBorders",
		GroupSetup			= L["By class"],
		GroupSetups			= {},
		Positions				= {},
		StatusMaps			= {},
		HideIn5Man			= false,
	})

	self:RegisterChatCommand("/srf", "/sraidframes", self.options)

	self.opt = self.db.profile

	self.OnMenuRequest = self.options

	-- Init variables
	self.enabled = false
	self.frames, self.groupframes = {}, {}, {}
	self.res, self.RangeChecks = {}, {}
	self.FramesByUnit = {}
	self.statusstate = {}
	self.MainTanks = {}

	self.cooldownSpells = {}
	self.cooldownSpells["WARLOCK"] = BS["Soulstone Resurrection"]
	self.cooldownSpells["DRUID"] = BS["Rebirth"]
	self.cooldownSpells["SHAMAN"] = BS["Reincarnation"]
	self.cooldownSpells["PALADIN"] = BS["Divine Intervention"]

	self.cleanseTypes= {
		["PRIEST"] = {
			["Magic"] = true,
			["Disease"] = true,
		},
		["SHAMAN"] = {
			["Poison"] = true,
			["Disease"] = true,
		},
		["PALADIN"] = {
			["Magic"] = true,
			["Poison"] = true,
			["Disease"] = true,
		},
		["MAGE"] = {
			["Curse"] = true,
		},
		["DRUID"] = {
			["Curse"] = true,
			["Poison"] = true,
		},
	}

	self:AddStatusMap("ReadyCheck_Pending", 80, {"background"}, L["Ready?"], {r = 0.1, g = 0.1, b = 0.1})

	self:AddStatusMap("Death", 70, {"background"}, L["Dead"], {r = 0.1, g = 0.1, b = 0.1, a = 1})
	
	self:AddStatusMap("Buff_".. NormalizeSpellName(BS["Spirit of Redemption"]), 65, {"statusbar"}, L["Dead"], {r=1,g=0,b=0,a=1})
	self:AddStatusMap("Buff_".. NormalizeSpellName(BS["Divine Intervention"]), 65, {"statusbar"}, L["Intervened"], {r=1,g=0,b=0,a=1})

	self:AddStatusMap("Aggro", 50, {"border", "statusbar"}, "Aggro", {r = 1, g = 0, b = 0})
	self:AddStatusMap("Target", 55, {"border"}, "Target", {r = 1, g = 0.75, b = 0})
	self:AddStatusMap("Heal", 60, {"border"}, "Inc. heal", {r = 0, g = 1, b = 0})

	self:AddStatusMap("Debuff_Curse", 55, {"background"}, "Cursed", self.opt.DebuffColors["Curse"])
	self:AddStatusMap("Debuff_Magic", 54, {"background"}, "Magic", self.opt.DebuffColors["Magic"])
	self:AddStatusMap("Debuff_Disease", 53, {"background"}, "Diseased", self.opt.DebuffColors["Disease"])
	self:AddStatusMap("Debuff_Poison", 52, {"background"}, "Poisoned", self.opt.DebuffColors["Poison"])

	self:AddStatusMap("Buff_".. NormalizeSpellName(BS["Shield Wall"]), 53, {"statusbar"}, BS["Shield Wall"], {r=1,g=1,b=1,a=1})
	self:AddStatusMap("Buff_".. NormalizeSpellName(BS["Last Stand"]), 52, {"statusbar"}, BS["Last Stand"], {r=1,g=1,b=1,a=1})
	self:AddStatusMap("Buff_".. NormalizeSpellName(BS["Vanish"]), 51, {"statusbar"}, L["Vanished"], {r=0,g=1,b=0,a=1})
	self:AddStatusMap("Buff_".. NormalizeSpellName(BS["Invisibility"]), 51, {"statusbar"}, BS["Invisibility"], {r=0,g=1,b=0,a=1})
	self:AddStatusMap("Buff_".. NormalizeSpellName(BS["Gift of Life"]), 51, {"statusbar"}, BS["Gift of Life"], {r=1,g=1,b=1,a=1})

	self:AddStatusMap("Buff_".. NormalizeSpellName(BS["Evasion"]), 50, {"statusbar"}, BS["Evasion"], {r=1,g=1,b=0,a=1})
	self:AddStatusMap("Buff_".. NormalizeSpellName(BS["Blessing of Protection"]), 50, {"statusbar"}, L["Protection"], {r=1,g=1,b=1,a=1})
	self:AddStatusMap("Buff_".. NormalizeSpellName(BS["Stealth"]), 50, {"statusbar"}, L["Stealthed"], {r=1,g=1,b=1,a=1})
	self:AddStatusMap("Buff_".. NormalizeSpellName(BS["Innervate"]), 51, {"statusbar"}, L["Innervating"], {r=0,g=1,b=0,a=1})
	self:AddStatusMap("Buff_".. NormalizeSpellName(BS["Ice Block"]), 50, {"statusbar"}, BS["Ice Block"], {r=1,g=1,b=1,a=1})
	self:AddStatusMap("Buff_".. NormalizeSpellName(BS["Feign Death"]), 50, {"statusbar"}, BS["Feign Death"], {r=0,g=1,b=0,a=1})
	self:AddStatusMap("Buff_".. NormalizeSpellName(BS["Cloak of Shadows"]), 50, {"statusbar"}, BS["Cloak of Shadows"], {r=1,g=1,b=1,a=1})
	self:AddStatusMap("Buff_".. NormalizeSpellName(BS["Divine Shield"]), 50, {"statusbar"}, BS["Divine Shield"], {r=1,g=1,b=1,a=1})
	self:AddStatusMap("Buff_".. NormalizeSpellName(BS["Power Infusion"]), 50, {"statusbar"}, L["Infused"], {r=1,g=1,b=1,a=1})

	self:AddStatusMap("Buff_".. NormalizeSpellName(BS["Misdirection"]), 45, {"statusbar"}, BS["Misdirection"], {r=0,g=1,b=0,a=1})
	self:AddStatusMap("Buff_".. NormalizeSpellName(BS["Intervene"]), 45, {"statusbar"}, BS["Intervene"], {r=0,g=1,b=0,a=1})
	self:AddStatusMap("Buff_".. NormalizeSpellName(BS["Fear Ward"]), 40, {"statusbar"}, BS["Fear Ward"], {r=1,g=1,b=0,a=1})
	self:AddStatusMap("Buff_".. NormalizeSpellName(BS["Heroism"]), 39, {"statusbar"}, BS["Heroism"], {r=1,g=1,b=1,a=1})
	self:AddStatusMap("Buff_".. NormalizeSpellName(BS["Bloodlust"]), 39, {"statusbar"}, BS["Bloodlust"], {r=1,g=1,b=1,a=1})
	self:AddStatusMap("Buff_".. NormalizeSpellName(BS["Shadowform"]), 35, {"statusbar"}, BS["Shadowform"], {r=1,g=0,b=0.75,a=1})

	self.statusElements = {["border"] = true, ["background"] = true, ["statusbar"] = true}
	
	self:DefaultGroupSetups()
	
	self:chatUpdateFilterMenu()
	self:chatUpdateBuffMenu()
	self:chatUpdateDebuffMenu()
	self:chatUpdateStatusElements()

	self.master = CreateFrame("Frame", "sRaidFrame", UIParent)
	self.master:SetMovable(true)
	self.master:SetScale(self.opt.Scale)

	self.master:SetHeight(200);
	self.master:SetWidth(200);
end

function sRaidFrames:OnProfileEnable()
	self.opt = self.db.profile
	self:SetPosition()
end

function sRaidFrames:OnEnable()
	self.PlayerClass = select(2, UnitClass("player"))

	self:AddRangeFunction(10, function (unit) return CheckInteractDistance(unit, 3) == 1 end)
	self:AddRangeFunction(28, function (unit) return CheckInteractDistance(unit, 4) == 1 end)
	self:AddRangeFunction(100, function (unit) return UnitIsVisible(unit) == 1 end)

	self:ScanSpellbookForRange()

	self:RegisterBucketEvent("RAID_ROSTER_UPDATE", 0.2, "UpdateRoster")
	self:RegisterEvent("LEARNED_SPELL_IN_TAB", "ScanSpellbookForRange")
	self:RegisterEvent("RosterLib_UnitChanged")

	self:UpdateRoster()
end

function sRaidFrames:OnDisable()
	self.enabled = false
	self.master:Hide()
end

local EventsHealth = {"UNIT_HEALTH", "UNIT_MAXHEALTH"}
local EventsPower = {"UNIT_MANA", "UNIT_RAGE", "UNIT_ENERGY", "UNIT_MAXMANA", "UNIT_MAXRAGE", "UNIT_MAXFOCUS", "UNIT_MAXENERGY", "UNIT_DISPLAYPOWER"}

function sRaidFrames:EnableFrames()
	self.enabled = true
	self.statusstate = {}

	self:CreateFrames()

	self:RegisterBucketEvent(EventsHealth, 0.05, "UNIT_HEALTH")
	self:RegisterBucketEvent(EventsPower, 1, "UNIT_POWER")
	self:RegisterBucketEvent("UNIT_AURA", 0.2)

	self:RegisterEvent("PLAYER_TARGET_CHANGED", "UpdateTarget")
	
	if HealComm then
		HealComm.RegisterCallback(self, "HealComm_DirectHealStart")
		HealComm.RegisterCallback(self, "HealComm_DirectHealStop")
		--HealComm.RegisterCallback(self, "HealComm_DirectHealDelayed")
	end

	self:RegisterEvent("Banzai_UnitGainedAggro")
	self:RegisterEvent("Banzai_UnitLostAggro")

	self:RegisterEvent("oRA_PlayerCanResurrect")
	self:RegisterEvent("oRA_PlayerResurrected")
	self:RegisterEvent("oRA_PlayerNotResurrected")
	self:RegisterEvent("oRA_MainTankUpdate")

	self:RegisterEvent("SpecialEvents_UnitDebuffGained")
	self:RegisterEvent("SpecialEvents_UnitDebuffLost")
	self:RegisterEvent("SpecialEvents_UnitBuffGained")
	self:RegisterEvent("SpecialEvents_UnitBuffLost")
	
	self:RegisterEvent("READY_CHECK")
	self:RegisterEvent("READY_CHECK_CONFIRM")
	self:RegisterEvent("READY_CHECK_FINISHED") 

	self:ScheduleRepeatingEvent("sRaidFramesRangeCheck", self.RangeCheck, self.opt.RangeFrequency, self)

	self:UpdateRoster()
	self:oRA_MainTankUpdate()

	self.master:Show()
end

function sRaidFrames:DisableFrames()
	if not self.enabled then return end
	self.statusstate = {}
	self.FramesByUnit = {}

	self:UnregisterBucketEvent(EventsHealth)
	self:UnregisterBucketEvent(EventsPower)
	self:UnregisterBucketEvent("UNIT_AURA")

	self:UnregisterEvent("PLAYER_TARGET_CHANGED")

	self:UnregisterEvent("Banzai_UnitGainedAggro")
	self:UnregisterEvent("Banzai_UnitLostAggro")

	self:UnregisterEvent("oRA_PlayerCanResurrect")
	self:UnregisterEvent("oRA_PlayerResurrected")
	self:UnregisterEvent("oRA_PlayerNotResurrected")
	self:UnregisterEvent("oRA_MainTankUpdate")

	self:UnregisterEvent("SpecialEvents_UnitDebuffGained")
	self:UnregisterEvent("SpecialEvents_UnitDebuffLost")
	self:UnregisterEvent("SpecialEvents_UnitBuffGained")
	self:UnregisterEvent("SpecialEvents_UnitBuffLost")

	self:CancelScheduledEvent("sRaidFramesRangeCheck")
	
	if HealComm then
		HealComm.UnregisterCallback(self, "HealComm_DirectHealStart")
		HealComm.UnregisterCallback(self, "HealComm_DirectHealStop")
		--HealComm.UnregisterCallback(self, "HealComm_HealModifierUpdate")
	end

	self.enabled = false
	self.master:Hide()
end

function sRaidFrames:UpdateRoster()
	local inRaid = UnitInRaid("player")

	if not inRaid and self.enabled then
		self:DisableFrames()
	end

	if inRaid and not self.enabled then
		if not (self.opt.HideIn5Man and GetNumRaidMembers() <= 5 ) then
			self:EnableFrames()
		end
	end
end

function sRaidFrames:UpdateAllUnits()
	for unit in pairs(self:GetAllUnits()) do
		self:UpdateAll(unit)
	end
end

function sRaidFrames:UpdateAll(unit)
	self:UpdateUnitDetails(unit)
	self:UpdateUnitHealth(unit)
	self:UpdateUnitPower(unit)
	self:UpdateStatuses(unit)
	self:UpdateAuras(unit)
end

function sRaidFrames:UNIT_POWER(units)
	for unit in pairs(units) do
		self:UpdateUnitPower(unit)
	end
end

function sRaidFrames:UNIT_HEALTH(units)
	for unit in pairs(units) do
		self:UpdateUnitHealth(unit)
	end
end

function sRaidFrames:UNIT_AURA(units)
	for unit in pairs(units) do
		self:UpdateAuras(unit)
	end
end

function sRaidFrames:READY_CHECK(author)
	if not self.opt.ReadyCheck or not (IsRaidLeader() or IsRaidOfficer()) then return end
	
	local authorid = roster:GetUnitIDFromName(author)
		
	for unitid in pairs(self:GetAllUnits()) do
		if unitid ~= authorid then
			self:SetStatus(unitid, "ReadyCheck_Pending")
		end
	end
end

function sRaidFrames:READY_CHECK_CONFIRM(id, confirm)
	if not self.opt.ReadyCheck then return end
	
	local unitid = "raid"..id
	
	self:UnsetStatus(unitid, "ReadyCheck_Pending")
	if confirm then
		self:SetStatus(unitid, "ReadyCheck_Ready")
	else
		self:SetStatus(unitid, "ReadyCheck_NotReady")
	end
end

function sRaidFrames:READY_CHECK_FINISHED()
	if not self.opt.ReadyCheck then return end
	
	for unitid in pairs(self:GetAllUnits()) do
		self:UnsetStatus(unitid, "ReadyCheck_Pending")
		self:UnsetStatus(unitid, "ReadyCheck_Ready")
		self:UnsetStatus(unitid, "ReadyCheck_NotReady")
	end
end

function sRaidFrames:RosterLib_UnitChanged(unitid, name, class, subgroup, rank, oldname, oldunitid, oldclass, oldsubgroup, oldrank)
	if not (unitid or oldunitid):match("^raid(%d+)$") then return end

	ShouldUpdateFrameCache = true

	if not name then
		-- Unit left
		self.statusstate[oldunitid] = nil
	elseif not oldname then
		-- Unit joined
		self.statusstate[unitid] = {}
		self:UpdateAll(unitid)
	else
		-- Unit changed
		self.statusstate[unitid] = {}
		self:UpdateAll(unitid)
	end

end

local LastTarget
function sRaidFrames:UpdateTarget()
	if not self.opt.HighlightTarget then return end

	if LastTarget then
		self:UnsetStatus(LastTarget, "Target")
	end

	for unit in pairs(self:GetAllUnits()) do
		if UnitIsUnit(unit, "target") then
			self:SetStatus(unit, "Target")
			LastTarget = unit
			break
		end
	end
end

function sRaidFrames:SpecialEvents_UnitDebuffGained(unit, debuff, apps, type, tex, rank, index)
	if not self:IsTracking(unit) then return end
	if not type or type == "" then return end
	if self.opt.DebuffFilter[debuff] then return end

	if self.opt.HighlightDebuffs == "never" then
		return -- Don't highlight it
	elseif self.opt.HighlightDebuffs == "onlyself" and not self:CanDispell(type) then
		return -- Don't highlight it
	end

	self:SetStatus(unit, "Debuff_".. type)
end

function sRaidFrames:SpecialEvents_UnitDebuffLost(unit, debuff, apps, type, tex, rank)
	if not self:IsTracking(unit) then return end
	if not type or type == "" then return end

	if not Aura:UnitHasDebuffType(unit, type) then
		self:UnsetStatus(unit, "Debuff_".. type)
	end
end

function sRaidFrames:SpecialEvents_UnitBuffGained(unit, buff, index, apps, tex, rank)
	if not self:IsTracking(unit) then return end

	if tex == "Interface\\Icons\\Spell_Nature_TimeStop" and buff == BS["Divine Intervention"] then
		self:SetStatus(unit, "Buff_"..NormalizeSpellName(buff))

	elseif tex == "Interface\\Icons\\Ability_Hunter_Misdirection" and buff == BS["Misdirection"] then
		self:SetStatus(unit, "Buff_"..NormalizeSpellName(buff))

	elseif tex == "Interface\\Icons\\Ability_Rogue_FeignDeath" and buff == BS["Feign Death"] then
		self:SetStatus(unit, "Buff_"..NormalizeSpellName(buff))

	elseif tex == "Interface\\Icons\\Ability_Warrior_VictoryRush" and buff == BS["Intervene"] then
		self:SetStatus(unit, "Buff_"..NormalizeSpellName(buff))

	elseif tex == "Interface\\Icons\\Spell_Nature_Lightning" and buff == BS["Innervate"] then
		self:SetStatus(unit, "Buff_"..NormalizeSpellName(buff))

	elseif tex == "Interface\\Icons\\Spell_Holy_GreaterHeal" and buff == BS["Spirit of Redemption"] then
		self:SetStatus(unit, "Buff_"..NormalizeSpellName(buff))
		
	elseif tex == "Interface\\Icons\\Ability_Warrior_ShieldWall" and buff == BS["Shield Wall"] then
		self:SetStatus(unit, "Buff_"..NormalizeSpellName(buff))

	elseif tex == "Interface\\Icons\\Spell_Holy_AshesToAshes" and buff == BS["Last Stand"] then
		self:SetStatus(unit, "Buff_"..NormalizeSpellName(buff))

	elseif tex == "Interface\\Icons\\INV_Misc_Gem_Pearl_05" and buff == BS["Gift of Life"] then
		self:SetStatus(unit, "Buff_"..NormalizeSpellName(buff))

	elseif tex == "Interface\\Icons\\Spell_Frost_Frost" and buff == BS["Ice Block"] then
		self:SetStatus(unit, "Buff_"..NormalizeSpellName(buff))

	elseif tex == "Interface\\Icons\\Spell_Holy_SealOfProtection" and buff == BS["Blessing of Protection"] then
		self:SetStatus(unit, "Buff_"..NormalizeSpellName(buff))

	elseif tex == "Interface\\Icons\\Spell_Holy_DivineIntervention" and buff == BS["Divine Shield"] then
		self:SetStatus(unit, "Buff_"..NormalizeSpellName(buff))

	elseif tex == "Interface\\Icons\\Ability_Vanish" and buff == BS["Vanish"] then
		self:SetStatus(unit, "Buff_"..NormalizeSpellName(buff))
		
	elseif tex == "Interface\\Icons\\Spell_Shadow_NetherCloak" and buff == BS["Cloak of Shadows"] then
		self:SetStatus(unit, "Buff_"..NormalizeSpellName(buff))

	elseif tex == "Interface\\Icons\\Ability_Mage_Invisibility" and buff == BS["Invisibility"] then
		self:SetStatus(unit, "Buff_"..NormalizeSpellName(buff))

	elseif tex == "Interface\\Icons\\Ability_Stealth" and buff == BS["Stealth"] then
		self:SetStatus(unit, "Buff_"..NormalizeSpellName(buff))

	elseif tex == "Interface\\Icons\\Spell_Shadow_ShadowWard" and buff == BS["Evasion"] then
		self:SetStatus(unit, "Buff_"..NormalizeSpellName(buff))

	elseif tex == "Interface\\Icons\\Spell_Holy_PowerInfusion" and buff == BS["Power Infusion"] then
		self:SetStatus(unit, "Buff_"..NormalizeSpellName(buff))
		
	elseif tex == "Interface\\Icons\\Ability_Shaman_Heroism" and buff == BS["Heroism"] then
		self:SetStatus(unit, "Buff_"..NormalizeSpellName(buff))
		
	elseif tex == "Interface\\Icons\\Spell_Nature_BloodLust" and buff == BS["Bloodlust"] then
		self:SetStatus(unit, "Buff_"..NormalizeSpellName(buff))

	elseif tex == "Interface\\Icons\\Spell_Holy_Excorcism" and buff == BS["Fear Ward"] then
		self:SetStatus(unit, "Buff_"..NormalizeSpellName(buff))

	elseif tex == "Interface\\Icons\\Spell_Shadow_Shadowform" and buff == BS["Shadowform"] then
		self:SetStatus(unit, "Buff_"..NormalizeSpellName(buff))
	end
end

function sRaidFrames:SpecialEvents_UnitBuffLost(unit, buff, apps, type, tex, rank)
	if not self:IsTracking(unit) then return end
	self:UnsetStatus(unit, "Buff_"..NormalizeSpellName(buff))
end

function sRaidFrames:Banzai_UnitGainedAggro(unit, unitTable)
	if not unit or not self:IsTracking(unit) or not self.opt.AggroCheck then return end

	self:SetStatus(unit, "Aggro")
end

function sRaidFrames:Banzai_UnitLostAggro(unit)
	if not unit or not self:IsTracking(unit) or not self.opt.AggroCheck then return end

	self:UnsetStatus(unit, "Aggro")
end

function sRaidFrames:oRA_PlayerCanResurrect(msg, player)
	local unit = roster:GetUnitIDFromName(player)
	if unit then
		self.res[unit] = 1
		self:UpdateUnitHealth(unit)
	end
end

function sRaidFrames:oRA_PlayerResurrected(msg, player)
	local unit = roster:GetUnitIDFromName(player)
	if unit then
		self.res[unit] = 2
		self:UpdateUnitHealth(unit)
	end
end

function sRaidFrames:oRA_PlayerNotResurrected(msg, player)
	local unit = roster:GetUnitIDFromName(player)
	if unit then
		self.res[unit] = nil
		self:UpdateUnitHealth(unit)
	end
end

function sRaidFrames:oRA_MainTankUpdate()
	if not oRA or not oRA.maintanktable then return end
	for i = 0, 10 do
		self.MainTanks[i] = oRA.maintanktable[i] or ""
	end
	
	self:ScheduleLeaveCombatAction(self.SetGroupFilters, self)
end

function sRaidFrames:IsUnitInRange(unit, range)
	if not self.RangeChecks[range] then
		return false
	end
	return self.RangeChecks[range](unit)
end

function sRaidFrames:CanDispell(type)
	return (self.cleanseTypes[self.PlayerClass] and self.cleanseTypes[self.PlayerClass][type])
end

function sRaidFrames:RangeCheck()
	if not self.opt.RangeCheck then return end
	local RangeLimit = self.opt.RangeLimit
	local RangeAlpha = self.opt.RangeAlpha

	for unit in pairs(self:GetAllUnits()) do
		if self:IsUnitInRange(unit, RangeLimit) then
			for _, f in pairs(self:FindUnitFrames(unit)) do
				f:SetAlpha(1)
			end
		else
			for _, f in pairs(self:FindUnitFrames(unit)) do
				f:SetAlpha(RangeAlpha)
			end
		end
	end
end

function sRaidFrames:AddRangeFunction(range, check)
	self.RangeChecks[range] = check
	self:UpdateRangeLimitOptions()
end

-- Credits to pastamancer <3
function sRaidFrames:ScanSpellbookForRange()
	local i = 1
	local Gratuity = AceLibrary("Gratuity-2.0")
	local sName, sRank, sRange

	repeat
		local sIndex = i
		sName, sRank = GetSpellName(i, BOOKTYPE_SPELL)
		if sName and IsSpellInRange(i, "spell", "player") ~= nil then
			Gratuity:SetSpell(i, BOOKTYPE_SPELL)
			sRange = select(3, Gratuity:Find(L["(%d+) yd range"], 2, 2))
			if sRange then
				self:AddRangeFunction(tonumber(sRange), function (unit) return IsSpellInRange(sIndex, "spell", unit) == 1 end)
			end
		end
		i = i + 1
	until not sName
end

function sRaidFrames:UpdateRangeFrequency()
		self:CancelScheduledEvent("sRaidFramesRangeCheck")
		self:ScheduleRepeatingEvent("sRaidFramesRangeCheck", self.RangeCheck, self.opt.RangeFrequency, self)
end

function sRaidFrames:UpdateStatuses(unit)
	for element in pairs(self.statusElements) do
		self:UpdateUnitStatusElement(unit, element)
	end
end

function sRaidFrames:UpdateUnitDetails(unit)
	if not self:IsTracking(unit) then return end
	for _, f in pairs(self:FindUnitFrames(unit)) do
		local class = select(2, UnitClass(unit))
		if class then
			f.title:SetText(UnitName(unit))

			local color = RAID_CLASS_COLORS[class]
			f.title:SetTextColor(color.r, color.g, color.b, 1)
		else
			f.title:SetText(UnitName(unit) or L["Unknown"])
		end
	end
end

function sRaidFrames:UpdateUnitHealth(unit)
	if not self:IsTracking(unit) then return end
	for _, f in pairs(self:FindUnitFrames(unit)) do
		local status, dead, ghost = nil, UnitIsDead(unit), UnitIsGhost(unit)

		if not UnitIsConnected(unit) then status = "|cffff0000"..L["Offline"].."|r"
		elseif self.res[unit] == 1 and dead then status = "|cffff8c00"..L["Can Recover"].."|r"
		elseif self.res[unit] == 2 and (dead or ghost) then status = "|cffff8c00"..L["Resurrected"].."|r"
		elseif ghost then status = "|cffff0000"..L["Released"].."|r"
		elseif dead then status = "|cffff0000"..L["Dead"].."|r"
		end

		if status then
			f.hpbar.text:SetText(status)
			f.hpbar:SetValue(0)
			f.mpbar:SetValue(0)
			self:SetStatus(unit, "Death")
		else
			self:UnsetStatus(unit, "Death")
			self.res[unit] = nil
			local hp = UnitHealth(unit) or 0
			local hpmax = UnitHealthMax(unit)
			local hpp = (hpmax ~= 0) and ceil((hp / hpmax) * 100) or 0
			local hptext, hpvalue = nil, nil

			local format = self.opt.HealthFormat
			if format == "percent" then
				hptext = hpp .."%"
			elseif format == "deficit" then
				hptext = (hp-hpmax) ~=  0 and (hp-hpmax) or nil
			elseif format == "current" then
				hptext = hp
			elseif format == "curmax" then
				hptext = hp .."/".. hpmax
			elseif format == "curdeficit" then
				hptext = (hp-hpmax) ~=  0 and  hp .." |cffff0000".. (hp-hpmax) or hpmax
			end

			if self.opt.Invert then
				hpvalue = 100 - hpp
			else
				hpvalue = hpp
			end

			f.hpbar.text:SetText(hptext)
			f.hpbar:SetValue(hpvalue)
			f.hpbar:SetStatusBarColor(self:GetHPSeverity(hpp/100))
		end
	end
end

function sRaidFrames:UpdateUnitPower(unit)
	if not self:IsTracking(unit) then return end
	for _, f in pairs(self:FindUnitFrames(unit)) do
		local powerType = UnitPowerType(unit)
		if self.opt.PowerFilter[powerType] == false then
			f.mpbar:SetValue(0)
		else
			local color = ManaBarColor[powerType]
			local mp = UnitMana(unit) or 0
			local mpmax = UnitManaMax(unit)
			local mpp = (mpmax ~= 0) and ceil((mp / mpmax) * 100) or 0
			f.mpbar:SetStatusBarColor(color.r, color.g, color.b)
			f.mpbar:SetValue(mpp)
		end
	end
end

function sRaidFrames:UpdateAuras(unit)
	if not self:IsTracking(unit) then return end
	local layout = self:GetLayout()
	for _, f in pairs(self:FindUnitFrames(unit)) do
		for i = 1, layout.debuffCount do
			f["aura".. i]:Hide()
		end

		for i = 1, layout.buffCount do
			f["buff".. i]:Hide()
		end

		local BuffType = self.opt.BuffType
		local DebuffSlots = 0
		if BuffType == "debuffs" or BuffType == "buffsifnotdebuffed" then
			local DebuffFilter = self.opt.DebuffFilter
			local DebuffWhitelist = self.opt.DebuffWhitelist
			local ShowOnlyDispellable = self.opt.ShowOnlyDispellable
			for i=1,40 do
				local debuffName, _, debuffTexture, debuffApplications, debuffType = UnitDebuff(unit, i)
				if not debuffTexture then break end
				if not DebuffFilter[debuffName] and ((ShowOnlyDispellable and (self:CanDispell(debuffType) or DebuffWhitelist[debuffName])) or not ShowOnlyDispellable) then
					DebuffSlots = DebuffSlots + 1
					local debuffFrame = f["aura".. DebuffSlots]
					debuffFrame.unitid = unit
					debuffFrame.debuffid = i
					debuffFrame.count:SetText(debuffApplications > 1 and debuffApplications or nil);
					debuffFrame.texture:SetTexture(debuffTexture)
					debuffFrame:Show()
				end

				if DebuffSlots == layout.debuffCount then break end
			end
		end

		if BuffType == "buffs" or (BuffType == "buffsifnotdebuffed" and DebuffSlots == 0) then
			local BuffSlots = 0
			local BuffFilter = self.opt.BuffFilter
			local showOnlyCastable = next(BuffFilter) ~= nil and 0 or 1
			for i=1,32 do
				local buffName, _, buffTexture, buffApplications = UnitBuff(unit, i, showOnlyCastable)
				if not buffTexture then break end

				if showOnlyCastable == 1 or BuffFilter[buffName] == true then
					BuffSlots = BuffSlots + 1
					local buffFrame = f["buff".. BuffSlots]
					buffFrame.buffid = i
					buffFrame.unitid = unit
					buffFrame.showCastable = showOnlyCastable
					buffFrame.count:SetText(buffApplications > 1 and buffApplications or nil)
					buffFrame.texture:SetTexture(buffTexture)
					buffFrame:Show()
				end

				if BuffSlots == layout.buffCount then break end
			end
		end
	end
end

function sRaidFrames:AddStatusMap(statuskey, priority, elements, text, color)
	--if self.opt.StatusMaps[statuskey] then return end
	
	if not color.a then color.a = 1 end
	
	self.opt.StatusMaps[statuskey] = {["priority"] = priority, ["elements"] = {}, ["text"] = text, ["color"] = color}
	for _, element in pairs(elements) do
		self.opt.StatusMaps[statuskey].elements[element] = true
	end
	self:chatUpdateStatusElements()
end

function sRaidFrames:SetStatus(unit, statuskey)
	if not self:IsTracking(unit) then return end

	local map = self.opt.StatusMaps[statuskey]

	if not map then return end
	
	if not self.statusstate[unit] then
		self.statusstate[unit] = {}
	end
	
	if self.statusstate[unit][statuskey] then return end
	
	self.statusstate[unit][statuskey] = map

	self:UpdateStatusElements(unit, statuskey)
end

function sRaidFrames:UnsetStatus(unit, statuskey)
	if not self:IsTracking(unit) then return end
	if self.opt.StatusMaps[statuskey] and self.statusstate[unit] and self.statusstate[unit][statuskey] then
		self.statusstate[unit][statuskey] = nil
		self:UpdateStatusElements(unit, statuskey)
	end
end

function sRaidFrames:UpdateStatusElements(unit, statuskey)
	for element in pairs(self.opt.StatusMaps[statuskey].elements) do
		self:UpdateUnitStatusElement(unit, element)
	end
end

function sRaidFrames:UpdateUnitStatusElement(unit, element)
	local status = self:GetTopStatus(element, unit)
	
	for _, f in pairs(self:FindUnitFrames(unit)) do
		if element == "border" then
			if status == nil then
				f:SetBackdropBorderColor(self.opt.BorderColor.r, self.opt.BorderColor.g, self.opt.BorderColor.b, self.opt.BorderColor.a)
			else
				f:SetBackdropBorderColor(status.color.r, status.color.g, status.color.b, status.color.a)
			end
		elseif element == "statusbar" then
			if status == nil then
				f.statustext:SetText(nil)
			else
				f.statustext:SetText(status.text)
				f.statustext:SetTextColor(status.color.r, status.color.g, status.color.b, status.color.a)
			end
		elseif element == "background" then
			if status == nil then
				f:SetBackdropColor(self.opt.BackgroundColor.r, self.opt.BackgroundColor.g, self.opt.BackgroundColor.b, self.opt.BackgroundColor.a)
			else
				f:SetBackdropColor(status.color.r, status.color.g, status.color.b, status.color.a)
			end
		end
	end
end

function sRaidFrames:GetTopStatus(element, unit)
	if not self.statusstate[unit] then
		return nil
	end

	local TopStatus, TopPriority = nil, 0
	for name, data in pairs(self.statusstate[unit]) do
		if self.opt.StatusMaps[name].elements[element] and self.opt.StatusMaps[name].priority > TopPriority then
			TopPriority = self.opt.StatusMaps[name].priority
			TopStatus = name
		end
	end

	return self.statusstate[unit][TopStatus] or nil
end

function sRaidFrames:GetHPSeverity(percent)
	if (percent >= 0.5) then
		return (1.0-percent)*2, 1.0, 0.0
	else
		return 1.0, percent*2, 0.0
	end
end

function sRaidFrames:QueryTooltipDisplay(value)
	if value == "never" then
		return false
	elseif value == "notincombat" and UnitAffectingCombat("player") then
		return false
	else
		return true
	end
end

function sRaidFrames:UnitTooltip(frame)
	local name, rank, subgroup, level, class, eclass, zone, _, _, role  = GetRaidRosterInfo(frame.id)
	local unit = frame:GetAttribute("unit")
	
	if not unit or not name then return end

	GameTooltip:SetOwner(frame)
	GameTooltip:AddDoubleLine(name, level > 0 and level or nil, RAID_CLASS_COLORS[eclass].r, RAID_CLASS_COLORS[eclass].g, RAID_CLASS_COLORS[eclass].b, 1, 1, 1)
	if UnitIsAFK(unit) then
		GameTooltip:AddLine(L["AFK: Away From Keyboard"], 1, 1, 0)
	end
	if UnitIsDND(unit) then
		GameTooltip:AddLine(L["DND: Do Not Disturb"], 1, 1, 0)
	end
	GameTooltip:AddLine(UnitRace(unit) .. " " .. class, 1, 1, 1);
	GameTooltip:AddDoubleLine(zone or UNKNOWN, L["Group %d"]:format(subgroup), 1, 1, 1, 1, 1, 1);

	if HealComm and not InCombatLockdown() then
		local version = HealComm:GetUnitVersion(unit)
		if version then
			GameTooltip:AddLine(("Tracking heals: v.%d"):format(version), 0, 1, 0)
		end
	end

	local cooldownSpell = self.cooldownSpells[eclass]
	if oRA and oRA:HasModule("OptionalCooldown") and cooldownSpell then
		if oRA:GetModule("OptionalCooldown").db.realm.cooldowns and oRA:GetModule("OptionalCooldown").db.realm.cooldowns[name] then
			local expire = oRA:GetModule("OptionalCooldown").db.realm.cooldowns[name]-time()
			if expire > 0 then
				GameTooltip:AddDoubleLine(cooldownSpell, SecondsToTime(expire), nil, nil, nil, 1, 0, 0)
			else
				GameTooltip:AddDoubleLine(cooldownSpell, L["Ready!"], nil, nil, nil, 0, 1, 0)
			end
		else
			GameTooltip:AddDoubleLine(cooldownSpell, UNKNOWN, nil, nil, nil, 1, 1, 0)
		end
	end

	GameTooltip:Show()
end

function sRaidFrames_initialHeaderConfigFunction(frame)
	local layout = sRaidFrames:GetLayout()
	frame:SetAttribute("initial-width", layout.unitframeWidth)
	frame:SetAttribute("initial-height", layout.unitframeHeight)
	frame:SetAttribute("type1", "target")
	frame:SetAttribute("*type1", "target")
end

function sRaidFrames_InitUnitFrame(frame)
	sRaidFrames:CreateUnitFrame(frame)
end

local ShouldUpdateFrameCache = false
function sRaidFrames_OnAttributeChanged(frame, name, value)
	if name == "unit" then
		ShouldUpdateFrameCache = true

		if value then
			frame.id = select(3, value:find("(%d+)"))
		end
	end
end

function sRaidFrames:UpdateFrameCache()
	self.FramesByUnit = {}
	for k, frame in pairs(self.frames) do
		local unit = frame:GetAttribute("unit")
		if unit then
			if not self.FramesByUnit[unit] then
				self.FramesByUnit[unit] = {}
			end
			tinsert(self.FramesByUnit[unit], frame)
		end
	end
	ShouldUpdateFrameCache = false
	self:UpdateAllUnits()
end

local function BuffFrame_OnEnter(this)
	if sRaidFrames:QueryTooltipDisplay(sRaidFrames.opt.BuffTooltipMethod) then
		GameTooltip:SetOwner(this)
		GameTooltip:SetUnitBuff(this.unitid, this.buffid, this.showCastable)
	end
end

local function BuffFrame_OnLeave(this)
	GameTooltip:Hide()
end

local function DebuffFrame_OnEnter(this)
	if sRaidFrames:QueryTooltipDisplay(sRaidFrames.opt.DebuffTooltipMethod) then
		GameTooltip:SetOwner(this)
		GameTooltip:SetUnitDebuff(this.unitid, this.debuffid)
	end
end

local function DebuffFrame_OnLeave(this)
	GameTooltip:Hide()
end

local DebuffFrame_OnLeave = BuffFrame_OnLeave

local function UnitFrame_OnEnter(this)
	if sRaidFrames:QueryTooltipDisplay(sRaidFrames.opt.UnitTooltipMethod) then
		sRaidFrames:UnitTooltip(this)
	end
end

local UnitFrame_OnLeave = BuffFrame_OnLeave


-- Adapts frames created by Secure Headers
function sRaidFrames:CreateUnitFrame(f)
	local layout = self:GetLayout()
	f:EnableMouse(true)
	f:RegisterForClicks("AnyUp")

	f:SetScript("OnEnter", UnitFrame_OnEnter)
	f:SetScript("OnLeave", UnitFrame_OnLeave)

	f.title = f:CreateFontString(nil, "ARTWORK")
	f.title:SetFontObject(GameFontNormalSmall)
	f.title:SetJustifyH("LEFT")


	for i = 1, layout.debuffCount do
		local debuffFrame = CreateFrame("Button", nil, f)
		debuffFrame:SetScript("OnEnter", DebuffFrame_OnEnter);
		debuffFrame:SetScript("OnLeave", DebuffFrame_OnLeave)
		debuffFrame.texture = debuffFrame:CreateTexture(nil, "ARTWORK")
		debuffFrame.texture:SetAllPoints(debuffFrame);
		debuffFrame.count = debuffFrame:CreateFontString(nil, "OVERLAY")
		debuffFrame.count:SetFontObject(GameFontHighlightSmallOutline)
		debuffFrame.count:SetJustifyH("CENTER")
		debuffFrame.count:SetPoint("CENTER", debuffFrame, "CENTER", 0, 0);
		debuffFrame:Hide()
		f["aura"..i] = debuffFrame
	end

	for i = 1, layout.buffCount do
		local buffFrame = CreateFrame("Button", nil, f)
		buffFrame:SetScript("OnEnter", BuffFrame_OnEnter)
		buffFrame:SetScript("OnLeave", BuffFrame_OnLeave)
		buffFrame.texture = buffFrame:CreateTexture(nil, "ARTWORK")
		buffFrame.texture:SetAllPoints(buffFrame)
		buffFrame.count = buffFrame:CreateFontString(nil, "OVERLAY")
		buffFrame.count:SetFontObject(GameFontHighlightSmallOutline)
		buffFrame.count:SetJustifyH("CENTER")
		buffFrame.count:SetPoint("CENTER", buffFrame, "CENTER", 0, 0);
		buffFrame:Hide()
		f["buff"..i] = buffFrame
	end

	local texture = media:Fetch(mType, self.opt.Texture)

	f.hpbar = CreateFrame("StatusBar", nil, f)
	f.hpbar:SetStatusBarTexture(texture)
	f.hpbar:SetMinMaxValues(0,100)
	f.hpbar:SetValue(0)

	f.hpbar.text = f.hpbar:CreateFontString(nil, "ARTWORK")
	f.hpbar.text:SetFontObject(GameFontHighlightSmall)
	f.hpbar.text:SetJustifyH("CENTER")

	f.mpbar = CreateFrame("StatusBar", nil, f)
	f.mpbar:SetStatusBarTexture(texture)
	f.mpbar:SetMinMaxValues(0,100)
	f.mpbar:SetValue(0)

	f.statustext = f.mpbar:CreateFontString(nil, "ARTWORK")
	f.statustext:SetFontObject(GameFontHighlightSmall)
	f.statustext:SetJustifyH("CENTER")

	layout.StyleUnitFrame(f)

	f:Hide();

	tinsert(self.frames, f)
	ShouldUpdateFrameCache = true

	ClickCastFrames = ClickCastFrames or {}
	ClickCastFrames[f] = true
end

function sRaidFrames:CreateFrames()
	if not self.enabled then return end

	local neededGroups = #self:GetCurrentGroupSetup()
	local createdGroups = #self.groupframes
	for i = createdGroups+1, neededGroups do
		self:CreateGroupFrame(i)
	end
	self:SetPosition()
	self:SetGroupFilters()
	self:SetGrowth()
end

function sRaidFrames:SetGroupFilters()
	if InCombatLockdown() then return end
	
	for _, f in pairs(self.groupframes) do
		local id = f:GetID()
		local filter = self:GetCurrentGroupSetup()[id]
		if filter then
			for attribute, value in pairs(filter.attributes) do
				f.header:SetAttribute(attribute, value or nil)
			end
			if filter.MagicMaintankList then
				f.header:SetAttribute("nameList", table.concat(self.MainTanks, ","))
			end
			f.title:SetText(filter.caption)
			f.header:Show()
		else
			f.header:Hide()
			f.header:SetAttribute("groupFilter", false)
			f.header:SetAttribute("nameList", false)
		end
		self:UpdateTitleVisibility(f.header)
	end
end

function sRaidFrames:GroupFrameGetNumChildren(frame)
	local i = 1
	local child = frame:GetAttribute("child"..i)
	while child do
		if not child:IsVisible() or not UnitExists(child:GetAttribute("unit")) then
			break
		end
		i = i + 1
		child = frame:GetAttribute("child"..i)
	end
	return (i-1)
end

function sRaidFrames:UpdateTitleVisibility(frame)
	if self.opt.ShowGroupTitles and self:GroupFrameGetNumChildren(frame) > 0 then
		frame:GetParent().anchor:Show()
	else
		frame:GetParent().anchor:Hide()
	end
end

function sRaidFrames:SetGrowth()
	for _, f in pairs(self.groupframes) do
		f.header:ClearAllPoints();
		if self.opt.Growth == "up" then
			f.header:SetAttribute("point", "TOP")
			f.header:SetPoint("BOTTOM", f.anchor, "TOP")
		elseif self.opt.Growth == "right" then
			f.header:SetAttribute("point", "LEFT")
			f.header:SetPoint("LEFT", f.anchor, "RIGHT")
		elseif self.opt.Growth == "left" then
			f.header:SetAttribute("point", "RIGHT")
			f.header:SetPoint("RIGHT", f.anchor, "LEFT")
		elseif self.opt.Growth == "down" then
			f.header:SetAttribute("point", "BOTTOM")
			f.header:SetPoint("TOP", f.anchor, "BOTTOM")
		end
	end
	self:SetSpacing()
end

function sRaidFrames:SetSpacing()
	local s = sRaidFrames.opt.Spacing
	for _, f in pairs(self.groupframes) do
		if self.opt.Growth == "down" then
			f.header:SetAttribute("xOffset", 0)
			f.header:SetAttribute("yOffset", s)
		elseif self.opt.Growth == "up" then
			f.header:SetAttribute("xOffset", 0)
			f.header:SetAttribute("yOffset", -s)
		elseif self.opt.Growth == "left" then
			f.header:SetAttribute("xOffset", s)
			f.header:SetAttribute("yOffset", 0)
		elseif self.opt.Growth == "right" then
			f.header:SetAttribute("xOffset", -s)
			f.header:SetAttribute("yOffset", 0)
		end
	end
end

function sRaidFrames:StartMovingAll(f)
	this.multidrag = 1
	local id = f:GetID()
	local fg = self.groupframes[id]
	local x, y = fg:GetLeft(), fg:GetTop()
	if ( not x or not y ) then
		return
	end
	for k, f in pairs(self.groupframes) do
		if k ~= id then
			local oX, oY = f:GetLeft(), f:GetTop()
			if ( oX and oY ) then
				f:ClearAllPoints()
				f:SetPoint("TOPLEFT", fg, "TOPLEFT", oX-x, oY-y)
			end
		end
	end
end

function sRaidFrames:StopMovingOrSizingAll(f)
	this.multidrag = nil
	local id = this:GetID()
	local fg = self.groupframes[id]
	for k, f in pairs(self.groupframes) do
		if k ~= id then
			local oX, oY = f:GetLeft(), f:GetTop()
			if ( oX and oY ) then
				f:ClearAllPoints()
				f:SetPoint("TOPLEFT", UIParent, "BOTTOMLEFT", oX, oY)
			end
		end
	end
end

local function sRaidFrames_OnGroupFrameEvent(frame, event)
	if event == "PARTY_MEMBERS_CHANGED" and frame:IsVisible() then
  	sRaidFrames:ScheduleLeaveCombatAction(sRaidFrames.UpdateTitleVisibility, sRaidFrames, frame)
	end
end

function sRaidFrames:CreateGroupFrame(id)
	local layout = self:GetLayout()
	local f = CreateFrame("Frame", "sRaidFramesGroupBase".. id, self.master)
	f:SetHeight(layout.headerHeight)
	f:SetWidth(layout.headerWidth)
	f:SetMovable(true)
	f:SetID(id)

	f.header = CreateFrame("Frame", "sRaidFramesGroupHeader" .. id, f, "SecureRaidGroupHeaderTemplate")
	f.header:SetAttribute("template", "sRaidFrameUnitSecure")
	f.header.initialConfigFunction = sRaidFrames_initialHeaderConfigFunction
	f.header:HookScript("OnEvent", sRaidFrames_OnGroupFrameEvent)
	
	-- hack to reduce login lockup times
	f.header:UnregisterEvent("UNIT_NAME_UPDATE")

	f.anchor = CreateFrame("Button", "sRaidFramesAnchor"..id, self.master)
	f.anchor:SetHeight(layout.headerHeight)
	f.anchor:SetWidth(layout.headerWidth)
	f.anchor:SetScript("OnDragStart", function() if self.opt.Locked then return end if IsAltKeyDown() then self:StartMovingAll(f) end f:StartMoving() end)
	f.anchor:SetScript("OnDragStop", function() if this.multidrag == 1 then self:StopMovingOrSizingAll() end f:StopMovingOrSizing(f) self:SavePosition() end)
	f.anchor:EnableMouse(true)
	f.anchor:RegisterForDrag("LeftButton")
	f.anchor:RegisterForClicks("AnyUp")
	f.anchor:SetPoint("CENTER", f)

	f.title = f.anchor:CreateFontString(nil, "OVERLAY", "GameFontNormal")
	f.title:SetJustifyH("CENTER")
	f.title:ClearAllPoints()
	f.title:SetPoint("CENTER", f.anchor)

	self.groupframes[id] = f
end

function sRaidFrames:SetWHP(frame, width, height, p1, relative, p2, x, y)
	frame:SetWidth(width)
	frame:SetHeight(height)

	if (p1) then
		frame:ClearAllPoints()
		frame:SetPoint(p1, relative, p2, x, y)
	end
end

function sRaidFrames:GetAllUnits()
	if ShouldUpdateFrameCache then self:UpdateFrameCache() end
	return self.FramesByUnit
end

function sRaidFrames:FindUnitFrames(unit)
	if ShouldUpdateFrameCache then self:UpdateFrameCache() end
	return self.FramesByUnit[unit] or {}
end

function sRaidFrames:IsTracking(unit)
	if ShouldUpdateFrameCache then self:UpdateFrameCache() end
	return (self.FramesByUnit[unit] ~= nil)
end

function sRaidFrames:SetPosition()
	if #self.opt.Positions <= 0 then
		self:ResetPosition()
	else
		self:RestorePosition()
	end
end

function sRaidFrames:SavePosition()
	for k,frame in pairs(self.groupframes) do
		local scale = frame:GetEffectiveScale()
		if scale then
			self.opt.Positions[k] = {x = frame:GetLeft()*scale, y = frame:GetTop()*scale}
		end
	end
end

function sRaidFrames:RestorePosition()
	local aryPos = self.opt.Positions
	local scale = self.master:GetEffectiveScale()

	for k, pos in pairs(aryPos) do
		local f = self.groupframes[k]
		if f then
			f:ClearAllPoints()
			f:SetPoint("TOPLEFT", UIParent, "BOTTOMLEFT", pos.x/scale, pos.y/scale)
		end
	end
end

function sRaidFrames:ResetPosition()
	self:PositionLayout("ctra", 200, -200)
end

function sRaidFrames:PositionLayout(layout, xBuffer, yBuffer)
	local xMod, yMod, i = 0, 0, -1
	local frameHeight = self:GetLayout().unitframeHeight+3+self.opt.Spacing
	local framePadding = MEMBERS_PER_RAID_GROUP

	for k,frame in pairs(self.groupframes) do
		i = i + 1
		if layout == "horizontal" then
			yMod = i * frame:GetWidth()
			xMod = 0
		elseif layout == "vertical" then
			if i ~= 0 and math.fmod(i, 2) == 0 then
				xMod = xMod + (-1*framePadding*frameHeight)
				yMod = 0
				i = 0
			else
				yMod = i * frame:GetWidth()
			end
		elseif layout == "ctra" then
			if i ~= 0 and math.fmod(i, 2) == 0 then
				yMod = yMod + frame:GetWidth()
				xMod = 0
				i = 0
			else
				xMod = i * (-1*framePadding*frameHeight)
			end
		end

		frame:ClearAllPoints()
		frame:SetPoint("TOPLEFT", UIParent, "TOPLEFT", xBuffer+yMod, yBuffer+xMod)
	end

	self:SavePosition()
end

local Tablet = AceLibrary("Tablet-2.0")
function sRaidFrames:OnTooltipUpdate()
	local cat = Tablet:AddCategory("columns", 2,
																 "text", L["Filter set"],
																 "justify", "LEFT",
																 "text2", L["State"],
																 "justify2", "LEFT",
																 "child_justify", "LEFT",
																 "child_justify2", "LEFT")
	local func, clicktext
	for name, _ in pairs(self:GetAllGroupFilters()) do
		if name == self.opt.GroupSetup then
			clicktext = L["|cff00ff00<Active>|r"]
			func = nil
		elseif InCombatLockdown() then
			clicktext = "|cffff0000<In Combat>|r"
			func = nil
		else
			clicktext = L["|cffffffff<Click to activate>|r"]
			func = self.SetCurrentGroupSetup
		end
	
		cat:AddLine("text", name,
							  "text2", clicktext,
							  "func", func,
							  "arg1", self,
							  "arg2", name)
	end
	--Tablet:SetHint(L["Right-click for options."])
end

function sRaidFrames:S(var, val)
	self.db.profile[var] = val
end
